/*
 * File Quota VFS module for Samba 3.0.24
 * Copyright (c) TAKEDA Yasuma, 2008
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * smb.conf format is here.
 *   vfs_objects = file_quota
 *   file_quota:quota = *.jpg:500k *.png:20k *.pdf:2M
 *   file_quota:exclude_users = administrator @admins
 *
 * These settings restricts a user can't create files over these limits.
 * And you can use "*" and "?" as wild card.
 */

#define VFS_FILE_QUOTA_VER 2008011701

#include "includes.h"

extern struct current_user current_user;

static int vfs_file_quota_debug_level = DBGC_VFS;
#undef DBGC_CLASS
#define DBGC_CLASS DBGC_VFS

/* Function prototypes */
static ssize_t file_quota_read(vfs_handle_struct *handle, files_struct *fsp, int fd, void *data, size_t n);

static ssize_t file_quota_pread(vfs_handle_struct *handle, files_struct *fsp, int fd, void *data, size_t n, SMB_OFF_T offset);

static ssize_t file_quota_write(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data, size_t n);

static ssize_t file_quota_pwrite(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data, size_t n, SMB_OFF_T offset);

static int file_quota_aio_read(struct vfs_handle_struct *handle, struct files_struct *fsp, SMB_STRUCT_AIOCB *aiocb);

static int file_quota_aio_write(struct vfs_handle_struct *handle, struct files_struct *fsp, SMB_STRUCT_AIOCB *aiocb);

static int file_quota_ftruncate(vfs_handle_struct *handle, files_struct *fsp, int fd, SMB_OFF_T len);

static int file_quota_rename(vfs_handle_struct *handle, connection_struct *conn, const char *oldname, const char *newname);

static int64 match_pattern_files(vfs_handle_struct *handle, const char *filename);
static BOOL has_exceeded_filesize(vfs_handle_struct *handle, const char *filename, int64 limit);
static BOOL will_exceed_filesize(vfs_handle_struct *handle, const char *filename, int64 limit, size_t len);
static int64 xstrtoi64(const char *ptr);

#define VFS_FILE_QUOTA_NOLIMIT -1

/* VFS operations structure */
static vfs_op_tuple file_quota_ops[] = {
#ifdef FILE_QUOTA_READ_OP
	{SMB_VFS_OP(file_quota_read), SMB_VFS_OP_READ, SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(file_quota_pread), SMB_VFS_OP_PREAD, SMB_VFS_LAYER_TRANSPARENT},
#endif
	{SMB_VFS_OP(file_quota_write), SMB_VFS_OP_WRITE, SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(file_quota_pwrite), SMB_VFS_OP_PWRITE, SMB_VFS_LAYER_TRANSPARENT},
#ifdef FILE_QUOTA_READ_OP
	{SMB_VFS_OP(file_quota_aio_read), SMB_VFS_OP_AIO_READ, SMB_VFS_LAYER_TRANSPARENT},
#endif
	{SMB_VFS_OP(file_quota_aio_write), SMB_VFS_OP_AIO_WRITE, SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(file_quota_ftruncate), SMB_VFS_OP_FTRUNCATE, SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(file_quota_rename), SMB_VFS_OP_RENAME, SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(NULL), SMB_VFS_OP_NOOP, SMB_VFS_LAYER_NOOP}
};

#ifdef FILE_QUOTA_READ_OP
/* Implementation of vfs_ops. */
static ssize_t file_quota_read(vfs_handle_struct *handle, files_struct *fsp, int fd, void *data, size_t n)
{
	ssize_t result;
	int64 limit;

	DEBUG(5, ("file_quota: read check [%s]\n", fsp->fsp_name));
	if (!fsp->is_directory){
		limit = match_pattern_files(handle, fsp->fsp_name);
		if (limit != VFS_FILE_QUOTA_NOLIMIT){
			if (has_exceeded_filesize(handle, fsp->fsp_name, limit) == True){
				DEBUG(3, ("file_quota: can not read [%s] because exceeded file size\n", fsp->fsp_name));
				return -1;
			}
		}
	}
	result = SMB_VFS_NEXT_READ(handle, fsp, fd, data, n);
	return result;
}


static ssize_t file_quota_pread(vfs_handle_struct *handle, files_struct *fsp, int fd, void *data, size_t n, SMB_OFF_T offset)
{
	int result;
	int64 limit, f_last;
	
	DEBUG(5, ("file_quota: pread check [%s]\n", fsp->fsp_name));

	if (!fsp->is_directory){
		limit = match_pattern_files(handle, fsp->fsp_name);
		if (limit != VFS_FILE_QUOTA_NOLIMIT){
			if (has_exceeded_filesize(handle, fsp->fsp_name, limit) == True){
				DEBUG(3, ("file_quota: can not pread [%s] because exceeded file size\n", fsp->fsp_name));
				return -1;
			}

			f_last = offset + n;
			if (will_exceed_filesize(handle, fsp->fsp_name, limit, f_last) == True){
				DEBUG(3, ("file_quota: can not pread [%s] because file size will exceed[%lld]\n", fsp->fsp_name, f_last));
				return -1;
			}
		}

	}
	result = SMB_VFS_NEXT_PREAD(handle, fsp, fd, data, n, offset);
	return result;
}
#endif

static ssize_t file_quota_write(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data, size_t n)
{
	ssize_t result;
	int64 limit;

	DEBUG(5, ("file_quota: write check [%s]\n", fsp->fsp_name));

	if (!fsp->is_directory){
		limit = match_pattern_files(handle, fsp->fsp_name);
		if (limit != VFS_FILE_QUOTA_NOLIMIT){
			if (has_exceeded_filesize(handle, fsp->fsp_name, limit) == True){
				DEBUG(3, ("file_quota: can not write [%s] because exceeded file size\n", fsp->fsp_name));
				return -1;
			}
		}
	}
	result = SMB_VFS_NEXT_WRITE(handle, fsp, fd, data, n);
	return result;
}


static ssize_t file_quota_pwrite(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data, size_t n, SMB_OFF_T offset)
{
	int result;
        int64 limit, f_last;

	DEBUG(5, ("file_quota: pwrite check [%s]\n", fsp->fsp_name));
	if (!fsp->is_directory){
		limit = match_pattern_files(handle, fsp->fsp_name);
		if (limit != VFS_FILE_QUOTA_NOLIMIT){
			if (has_exceeded_filesize(handle, fsp->fsp_name, limit) == True){
				DEBUG(3, ("file_quota: can not pwrite [%s] because file size has exceeded.\n", fsp->fsp_name));
				return -1;
			}

			f_last = offset + n;
			if (will_exceed_filesize(handle, fsp->fsp_name, limit, f_last) == True){
				DEBUG(3, ("file_quota: can not pwrite [%s] because file size will exceed[%lld]\n", fsp->fsp_name, f_last));
				return -1;
			}

		}
	}
	result = SMB_VFS_NEXT_PWRITE(handle, fsp, fd, data, n, offset);
	return result;
}

#ifdef FILE_QUOTA_READ_OP
static int file_quota_aio_read(struct vfs_handle_struct *handle, struct files_struct *fsp, SMB_STRUCT_AIOCB *aiocb)
{
	int result = 0;
	int64 limit, f_last;
	DEBUG(5, ("file_quota: aio read check [%s]\n", fsp->fsp_name));
	if (!fsp->is_directory){
		limit = match_pattern_files(handle, fsp->fsp_name);
		if (limit != VFS_FILE_QUOTA_NOLIMIT){
			if (has_exceeded_filesize(handle, fsp->fsp_name, limit) == True){
				DEBUG(3, ("file_quota: can not aio read [%s] because exceeded file size\n", fsp->fsp_name));
				return -1;
			}
		}
	}
	result = SMB_VFS_NEXT_AIO_READ(handle, fsp, aiocb);
	return result;
}
#endif

static int file_quota_aio_write(struct vfs_handle_struct *handle, struct files_struct *fsp, SMB_STRUCT_AIOCB *aiocb)
{
	int result = 0;
	int64 limit;

	DEBUG(5, ("file_quota: aio write check [%s]\n", fsp->fsp_name));

	if (!fsp->is_directory){
		limit = match_pattern_files(handle, fsp->fsp_name);
		if (limit != VFS_FILE_QUOTA_NOLIMIT){
			if (has_exceeded_filesize(handle, fsp->fsp_name, limit) == True){
				DEBUG(3, ("file_quota: can not aio write [%s] because exceeded file size\n", fsp->fsp_name));
				return -1;
			}
		}
	}
	result = SMB_VFS_NEXT_AIO_WRITE(handle, fsp, aiocb);
	return result;
}

static int file_quota_ftruncate(vfs_handle_struct *handle, files_struct *fsp, int fd, SMB_OFF_T len)
{
	int result = 0;
	int64 limit;

	DEBUG(5, ("file_quota: ftruncate check [%s]\n",fsp->fsp_name));
	limit = match_pattern_files(handle, fsp->fsp_name);
	if (limit != VFS_FILE_QUOTA_NOLIMIT){
		if (will_exceed_filesize(handle, fsp->fsp_name, limit, len) == True){
			DEBUG(3, ("file_quota: can not ftruncate [%s] because file size will exceed[%lld]\n", fsp->fsp_name, len));
			return -1;
		}
	}

	result = SMB_VFS_NEXT_FTRUNCATE(handle, fsp, fd, len);
	return result;
}

static int file_quota_rename(vfs_handle_struct *handle, connection_struct *conn, const char *oldname, const char *newname)
{
	int result = 0;
	int64 limit;
	
	DEBUG(5, ("file_quota: rename check [%s to %s]\n", oldname, newname));

	limit = match_pattern_files(handle, oldname);
	if (limit != VFS_FILE_QUOTA_NOLIMIT){
		if (has_exceeded_filesize(handle, oldname, limit) == True){
			DEBUG(3, ("file_quota: can not rename [%s(old)] because file size has exceeded.\n", oldname));
			errno = EACCES; /* permission denied */
			return -1;
		}
	}
	limit = match_pattern_files(handle, newname);
	if (limit != VFS_FILE_QUOTA_NOLIMIT){
		if (has_exceeded_filesize(handle, oldname, limit) == True){
			DEBUG(3, ("file_quota: can not rename [%s(new)] because file size has exceeded.\n", newname));
			errno = EACCES; /* permission denied */
			return -1;
		}

	}
	result = SMB_VFS_NEXT_RENAME(handle, conn, oldname, newname);
	return result;
}

NTSTATUS vfs_file_quota_init(void){
	NTSTATUS ret = smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "file_quota", file_quota_ops);
	
	if (!NT_STATUS_IS_OK(ret))
		return ret;

	vfs_file_quota_debug_level == debug_add_class("file_quota");
	if (vfs_file_quota_debug_level == -1){
		vfs_file_quota_debug_level = DBGC_VFS;
		DEBUG(0, ("%s: Couldn't register custom debugging class!\n", "vfs_file_quota_init"));
	} else {
		DEBUG(10, ("%s: Debug class number of '%s': %d\n", "vfs_file_quota_init","file_quota", vfs_file_quota_debug_level));
	}
	return ret;
}

/*
 * Parse and check the filename whether match or not
 * You can use "*" and "?" as wild card in the parameter. 
 * Sample format is here.
 * file_quota:quota=*.jpg:100k *.png:200k *.*:1m *:1g
 *
 * @return The size limit(includes 0) if matched.
 *         Negative value(-1) is not matched.
 */

static int64 match_pattern_files(vfs_handle_struct *handle, const char *filename){
	BOOL ret = False;
	const char **lp_list, **user_list;
	char *pattern, *pos;
	int64 limit = VFS_FILE_QUOTA_NOLIMIT;



	lp_list = lp_parm_string_list(SNUM(handle->conn), "file_quota", "quota", NULL);
	if (lp_list == NULL){
	  return limit;
	}

	/* check pattern match */
	for (;*lp_list;*lp_list++){
		pattern = SMB_STRDUP(*lp_list);

		pos = strchr_m(pattern, ':');

		if (!pos){
			/* wrong format because ':' is not included. */
			SAFE_FREE(pattern);
			continue;
		}

		*pos = '\0'; /* separate pattern and limit by NULL terminate */

		if (unix_wild_match(pattern, filename)){
			/* pattern matched */
		  
			/* check exuclude_users */
			user_list = lp_parm_string_list(SNUM(handle->conn), "file_quota", "exclude_users", NULL);
			if (user_list){
				if (token_contains_name_in_list(handle->conn->user, NULL, current_user.nt_user_token, user_list)){
					DEBUG(5, ("file_quota: exclude user :%s\n", handle->conn->user));
					return limit; /* limit is NOLIMIT at here */
				}
			}
			/* convert from string to integer and set the limit */
			limit = xstrtoi64(++pos);
			DEBUG(5, ("file_quota: filename [%s] is matched. pattern:[%s] limit %lld\n", filename, pattern, limit));
			SAFE_FREE(pattern);
			return limit;
		}
		SAFE_FREE(pattern);
	}

	return limit;
}

static BOOL has_exceeded_filesize(vfs_handle_struct *handle, const char *filename, int64 limit){
	SMB_STRUCT_STAT st;
	int ret;
	DEBUG(5, ("file_quota: filename [%s]/limit [%lld]\n", filename, limit));
        if (SMB_VFS_NEXT_STAT(handle, handle->conn, filename, &st) == 0) {
		if (limit < st.st_size) {
			DEBUG(3, ("file_quota: %s [size:%lld] has exceeded the limit [%lld].\n",filename, st.st_size, limit));
			return True;
		}
        }
	return False;
}

static BOOL will_exceed_filesize(vfs_handle_struct *handle, const char *filename, int64 limit, size_t len){

	SMB_STRUCT_STAT st;

	DEBUG(5, ("file_quota: filename [%s]/limit [%lld]/len %lld\n", filename, limit, len));
        if (SMB_VFS_NEXT_STAT(handle, handle->conn, filename, &st) == 0) {
		if (limit < st.st_size + len) {
			DEBUG(3, ("file_quota: %s[%lld] has exceeded the limit[%lld].\n",filename, st.st_size, limit));
			return True;
		}
        }
	return False;
}


/* 
 * more friendly strtoll function.
 * 1k/1K -> 1 * 1024
 * 1m/1M -> 1 * 1024 * 1024
 * 1g/1G -> 1 * 1024 * 1024 * 1024
 * 1t/1T, 1p/1P and so on.
 * @return int64 integer from string
 */
static int64 xstrtoi64 (const char *ptr)
{
	char *retptr;
	int64 ret = -1;
	/* bit width of 'signed' int64 */
	int ll_width = sizeof(int64) * 8 - 1;

	ret = (int64)strtoll (ptr, &retptr, 10);

	switch (*retptr) {
	case 'Y':
	case 'y':
		if ((ll_width > 80) && 
		    ((long long) ret < (long long)1<<(ll_width - 80)))
		{
			ret <<= 10;
		} else {
			ret = -1;
			break;
		}
	case 'Z':
	case 'z':
		if ((ll_width > 70) && 
		    ((long long) ret < (long long)1<<(ll_width - 70)))
		{
			ret <<= 10;
		} else {
			ret = -1;
			break;
		}
	case 'E':
	case 'e':
		if ((long long) ret < (long long)1<<(ll_width - 60)){
			ret <<= 10;
		} else {
			ret = -1;
			break;
		}
	case 'P':
	case 'p':
		if ((long long) ret < (long long)1<<(ll_width - 50)){
			ret <<= 10;
		} else {
			ret = -1;
			break;
		}
	case 'T':
	case 't':
		if ((long long) ret < (long long)1<<(ll_width - 40)){
			ret <<= 10;
		} else {
			ret = -1;
			break;
		}
	case 'G':
	case 'g':
		if ((long long) ret < (long long)1<<(ll_width - 30)){
			ret <<= 10;
		} else {
			ret = -1;
			break;
		}
	case 'M':
	case 'm':
		if ((long long) ret < (long long)1<<(ll_width - 20)){
			ret <<= 10;
		} else {
			ret = -1;
			break;
		}
	case 'K':
	case 'k':
		if ((long long) ret < (long long)1<<(ll_width - 10)){
			ret <<= 10;

		} else {
			ret = -1;
			break;
		}
		retptr++;
	default:
		break;
	}

	if (retptr && strlen(retptr)){
		/* wrong format */
		DEBUG(3, ("convert failed from %s to int64\n", ptr));
		ret = -1;
	}
	DEBUG(3, ("converted %s to %lld\n", ptr, ret));
	return ret;
}
